import argparse
import os
import random
import logging
import shutil
import time
import warnings
import numpy as np
import pprint
import math

import torch
import torch.nn as nn
import torch.nn.parallel
import torch.distributed as dist
import torch.optim
import torch.multiprocessing as mp
import torch.utils.data
import torch.utils.data.distributed
import torch.nn.functional as F

from datasets.cifar10 import CIFAR10_LT
from datasets.cifar100 import CIFAR100_LT
from datasets.places import Places_LT
from datasets.imagenet import ImageNet_LT
from datasets.ina2018 import iNa2018

import models
from losses import *

from utils import imb_create_logger
from utils import AverageMeter, ProgressMeter
from utils import accuracy, calibration

from methods import mixup_data, mixup_criterion
from methods import LabelAwareSmoothing, LearnableWeightScaling

model_names = sorted(name for name in models.__dict__
                     if name.islower() and not name.startswith("__")
                     and callable(models.__dict__[name]))

def parse_args():
    parser = argparse.ArgumentParser(description='One Stage Learning')
    parser.add_argument('--two_stage', default=False, action='store_false')
    parser.add_argument('--cfg', help='experiment configure file name', default='./config/cifar100/main_imb001.yaml', type=str)
    parser.add_argument('--data_path', default='../database/cifar100', type=str)
    parser.add_argument('--dataset', default='cifar100', help='dataset setting')
    parser.add_argument('--num_classes', default=100, type=int)
    parser.add_argument('-a', '--arch', metavar='ARCH', default='cifar_resnet32', choices=model_names, help='model architecture: ' + ' | '.join(model_names) + ' (default: resnet32)')
    parser.add_argument('--loss', default='LDAM', type=str, help='loss type')
    parser.add_argument('--imb_type', default='exp', type=str, help='imbalance type')
    parser.add_argument('--imb_factor', default=0.01, type=float, help='imbalance factor')
    parser.add_argument('--train_rule', default='DRW', type=str, help='data sampling strategy for train loader')
    parser.add_argument('--rand_number', default=0, type=int, help='fix random number for data sampling')
    parser.add_argument('--exp_str', default='1', type=str, help='number to indicate which experiment it is')
    parser.add_argument('-j', '--workers', default=4, type=int, metavar='N',
                        help='number of data loading workers (default: 4)')
    parser.add_argument('--num_epochs', default=200, type=int, metavar='N', help='number of total epochs to run')
    parser.add_argument('--start-epoch', default=0, type=int, metavar='N',
                        help='manual epoch number (useful on restarts)')
    parser.add_argument('--num_feats', default=128, type=int)
    parser.add_argument('--last_relu', default=False, type=bool)
    parser.add_argument('-b', '--batch-size', default=128, type=int, metavar='N', help='mini-batch size')
    parser.add_argument('--cos', default=False, action='store_true', help='if use cosine scheduler')
    parser.add_argument('--lr', '--learning-rate', default=0.1, type=float, metavar='LR', help='initial learning rate', dest='lr')
    parser.add_argument('--momentum', default=0.9, type=float, metavar='M', help='momentum')
    parser.add_argument('--wd', '--weight-decay', default=5e-4, type=float, metavar='W',
                        help='weight decay (default: 5e-4)', dest='weight_decay')
    parser.add_argument('-p', '--print-freq', default=10, type=int, metavar='N', help='print frequency (default: 10)')
    parser.add_argument('--load', default=False, type=bool)
    parser.add_argument('--resume', default='', type=str, metavar='PATH',
                        help='path to latest checkpoint (default: none)')
    parser.add_argument('-e', '--evaluate', dest='evaluate', action='store_true',
                        help='evaluate model on validation set')
    # parser.add_argument('--pretrained', dest='pretrained', action='store_true', help='use pre-trained model')

    parser.add_argument('--deterministic', default=True, type=bool)
    parser.add_argument('--seed', default=0, type=int, help='seed for initializing training. ')
    parser.add_argument('--dist_url', default='tcp://224.66.41.62:23456')
    parser.add_argument('--dist_backend', default='nccl')
    parser.add_argument('--multiprocessing_distributed', default=False, type=bool)
    parser.add_argument('--world_size', default=-1, type=int)
    parser.add_argument('--rank', default=-1, type=int)
    parser.add_argument('--distributed', default=False, type=bool)
    parser.add_argument('--gpu', default=0, type=int, help='GPU id to use.')

    parser.add_argument('--log_dir', type=str, default='log')
    parser.add_argument('--model_dir', type=str, default='checkpoint')
    parser.add_argument('--mode', type=str, default='')
    parser.add_argument('--s', type=float, default=1.5)
    parser.add_argument('--scale', type=float, default=30)
    parser.add_argument('--mixup', type=bool, default=False)
    parser.add_argument('--alpha', type=float, default=1.0)
    parser.add_argument('opts', help="Modify config options using the command-line", default=None,
                        nargs=argparse.REMAINDER)

    args = parser.parse_args()
    return args

best_acc1 = 0
its_ece = 100

def main():
    args = parse_args()
    config, logger = imb_create_logger(args)
    logger.info('\n' + pprint.pformat(args))
    logger.info('\n' + str(config))

    if config.deterministic:
        seed = config.seed
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        random.seed(seed)
        np.random.seed(seed)
        os.environ['PYTHONHASHSEED'] = str(seed)
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)

    if config.gpu is not None:
        warnings.warn('You have chosen a specific GPU, This will completely disable data parallelism.')

    if config.dist_url == "env://" and config.world_size == -1:
        config.world_size = int(os.environ["WORLD_SIZE"])

    config.distributed = config.world_size > 1 or config.multiprocessing_distributed

    ngpus_per_node = torch.cuda.device_count()
    if config.multiprocessing_distributed:
        config.world_size = ngpus_per_node * config.world_size
        mp.spawn(main_worker, nprocs=ngpus_per_node, args=(ngpus_per_node, config, logger))
    else:
        main_worker(config.gpu, ngpus_per_node, config, logger)

def main_worker(gpu, ngpus_per_node, config, logger):
    global best_acc1, its_ece
    config.gpu = gpu
    if config.gpu is not None:
        logger.info("Use GPU: {} for training".format(config.gpu))

    if config.distributed:
        if config.dist_url == "env://" and config.rank == -1:
            config.rank = int(os.environ["RANK"])
        if config.multiprocessing_distributed:
            # For multiprocessing distributed training, rank needs to be the
            # global rank among all the processes
            config.rank = config.rank * ngpus_per_node + gpu
        dist.init_process_group(backend=config.dist_backend, init_method=config.dist_url, world_size=config.world_size, rank=config.rank)

    if config.mode in ['init', 'fix']:
        path = './weights/weight1' if config.last_relu else './weights/weight0'
        weight = torch.Tensor(np.load(path + '_%dx%d.npy' % (config.num_classes, config.num_feats))).cuda(config.gpu)
    else:
        weight = None
    if config.mode == 'none':
        config.mode = ''
    if config.loss not in ['CE', 'BSoftmax', 'Focal']:
        mode = config.mode + 'norm'
    else:
        mode = config.mode + 'linear'

    model = models.create_model(config, weight, mode, two_stage=False)

    if not torch.cuda.is_available():
        logger.info('using CPU, this will be slow')
    elif config.distributed:
        if config.gpu is not None:
            torch.cuda.set_device(config.gpu)
            model.cuda(config.gpu)
            config.batch_size = int(config.batch_size / ngpus_per_node)
            config.workers = int((config.workers + ngpus_per_node - 1) / ngpus_per_node)
            model = torch.nn.parallel.DistributedDataParallel(model, device_ids=[config.gpu])
        else:
            model.cuda()
            model = torch.nn.parallel.DistributedDataParallel(model)
    elif config.gpu is not None:
        torch.cuda.set_device(config.gpu)
        model = model.cuda(config.gpu)
    else:
        model = torch.nn.DataParallel(model).cuda()

    if config.load:
        if os.path.isfile(config.resume):
            logger.info("=> loading checkpoint '{}'".format(config.resume))
            if config.gpu is None:
                checkpoint = torch.load(config.resume)
            else:
                # Map model to be loaded to specified single gpu.
                loc = 'cuda:{}'.format(config.gpu)
                checkpoint = torch.load(config.resume, map_location=loc)
            # config.start_epoch = checkpoint['epoch']
            best_acc1 = checkpoint['best_acc1']
            if config.gpu is not None:
                # best_acc1 may be from a checkpoint from a different GPU
                best_acc1 = best_acc1.to(config.gpu)
            model.load_state_dict(checkpoint['state_dict_model'])
            logger.info("=> loaded checkpoint '{}' (epoch {})"
                        .format(config.resume, checkpoint['epoch']))
        else:
            logger.info("=> no checkpoint found at '{}'".format(config.resume))

    # Data loading code
    if config.dataset == 'cifar10':
        dataset = CIFAR10_LT(config.distributed, root=config.data_path, imb_factor=config.imb_factor,
                             batch_size=config.batch_size, num_works=config.workers)

    elif config.dataset == 'cifar100':
        dataset = CIFAR100_LT(config.distributed, root=config.data_path, imb_factor=config.imb_factor,
                              batch_size=config.batch_size, num_works=config.workers)

    elif config.dataset == 'places':
        dataset = Places_LT(config.distributed, root=config.data_path,
                            batch_size=config.batch_size, num_works=config.workers)

    elif config.dataset == 'imagenet':
        dataset = ImageNet_LT(config.distributed, root=config.data_path,
                              batch_size=config.batch_size, num_works=config.workers)

    elif config.dataset == 'ina2018':
        dataset = iNa2018(config.distributed, root=config.data_path,
                          batch_size=config.batch_size, num_works=config.workers)
    else:
        raise ValueError('The dataset {} is not in list'.format(config.dataset))

    optimizer = torch.optim.SGD(model.parameters(), config.lr, momentum=config.momentum, weight_decay=config.weight_decay)

    train_loader = dataset.train_instance
    val_loader = dataset.eval
    cls_num_list = dataset.cls_num_list

    if config.distributed:
        train_sampler = dataset.dist_sampler
    for epoch in range(config.num_epochs):
        if config.distributed:
            train_sampler.set_epoch(epoch)

        if config.train_rule == 'Reweight':
            beta = 0.9999
            effective_num = 1.0 - np.power(beta, cls_num_list)
            per_cls_weights = (1.0 - beta) / np.array(effective_num)
            per_cls_weights = per_cls_weights / np.sum(per_cls_weights) * len(cls_num_list)
            per_cls_weights = torch.FloatTensor(per_cls_weights).cuda(config.gpu)
        elif config.train_rule == 'DRW':
            idx = epoch // 160
            betas = [0, 0.9999]
            effective_num = 1.0 - np.power(betas[idx], cls_num_list)
            per_cls_weights = (1.0 - betas[idx]) / np.array(effective_num)
            per_cls_weights = per_cls_weights / np.sum(per_cls_weights) * len(cls_num_list)
            per_cls_weights = torch.FloatTensor(per_cls_weights).cuda(config.gpu)
        else:
            per_cls_weights = None

        if config.loss == 'CE':
            criterion = nn.CrossEntropyLoss(weight=per_cls_weights).cuda(config.gpu)
        elif config.loss == 'LDAM':
            criterion = LDAMLoss(cls_num_list=cls_num_list, max_m=0.5, s=config.scale, weight=per_cls_weights).cuda(config.gpu)
        elif config.loss == 'Focal':
            criterion = FocalLoss(gamma=1).cuda(config.gpu)
        elif config.loss == 'LNorm':
            criterion = LMSoftmaxLoss(scale=config.scale, weight=per_cls_weights).cuda(config.gpu)
        elif config.loss == 'Norm':
            criterion = NormFaceLoss(scale=config.scale, weight=per_cls_weights).cuda(config.gpu)
        elif config.loss == 'CosFace':
            criterion = CosFaceLoss(scale=config.scale, weight=per_cls_weights).cuda()
        elif config.loss == 'ArcFace':
            criterion = ArcFaceLoss(scale=config.scale, weight=per_cls_weights).cuda()
        elif config.loss == 'SaMargin':
            criterion = SampleMarginLoss(weight=per_cls_weights)
        elif config.loss == 'BSoftmax':
            criterion = BalancedSoftmaxLoss(freq=cls_num_list).cuda()
        else:
            warnings.warn('Loss type is not listed')
            return
        adjust_learning_rate(optimizer, epoch, config)
        train(train_loader, model, criterion, optimizer, epoch, config, logger)

        acc1, ece = validate(val_loader, model, criterion, config, logger)

        is_best = acc1 > best_acc1
        best_acc1 = max(acc1, best_acc1)
        if is_best:
            its_ece = ece
        logger.info('Best Prec@1: %.3f%% ECE: %.3f%%\n' % (best_acc1, its_ece))
        if not config.multiprocessing_distributed or (config.multiprocessing_distributed
                                                      and config.rank % config.ngpus_per_node == 0):
            os.makedirs(config.model_dir, exist_ok=True)
            save_checkpoint({
                'epoch': epoch + 1,
                'state_dict_model': model.state_dict(),
                'best_acc1': best_acc1,
                'its_ece': its_ece,
            }, is_best, config.model_dir)


def train(train_loader, model, criterion, optimizer, epoch, config, logger):
    batch_time = AverageMeter('Time', ':6.3f')
    data_time = AverageMeter('Data', ':6.3f')
    losses = AverageMeter('Loss', ':.3f')
    top1 = AverageMeter('Acc@1', ':6.3f')
    top5 = AverageMeter('Acc@5', ':6.3f')
    margins = AverageMeter('ClsMargin', ':.3f')
    norms = AverageMeter('Norm', ':.3f')
    ratios = AverageMeter('Ratio', ':.3f')
    progress = ProgressMeter(len(train_loader), [batch_time, losses, top1, top5, margins, norms, ratios], prefix="Epoch: [{}]".format(epoch))

    model.train()
    end = time.time()
    for i, (images, target) in enumerate(train_loader):
        data_time.update(time.time() - end)
        if torch.cuda.is_available():
            images = images.cuda(config.gpu, non_blocking=True)
            target = target.cuda(config.gpu, non_blocking=True)

        if config.mixup:
            images, targets_a, targets_b, lam = mixup_data(images, target, alpha=config.stage1_alpha)
        output = config.s * model(images)
        loss = criterion(output, target) if not config.mixup else mixup_criterion(criterion, output, targets_a, targets_b, lam)

        margin, norm, ratio = model.margin()
        acc1, acc5 = accuracy(output, target, topk=(1, 5))
        losses.update(loss.item(), images.size(0))
        top1.update(acc1[0], images.size(0))
        top5.update(acc5[0], images.size(0))
        margins.update(margin)
        norms.update(norm)
        ratios.update(ratio)

        # compute gradient and do SGD step
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        batch_time.update(time.time() - end)
        end = time.time()

        if i % config.print_freq == 0:
            progress.display(i, logger)


def validate(val_loader, model, criterion, config, logger):
    batch_time = AverageMeter('Time', ':6.3f')
    losses = AverageMeter('Loss', ':.3f')
    top1 = AverageMeter('Acc@1', ':6.2f')
    top5 = AverageMeter('Acc@5', ':6.2f')

    progress = ProgressMeter(
        len(val_loader),
        [batch_time, losses, top1, top5],
        prefix='Eval: ')

    model.eval()
    class_num = torch.zeros(config.num_classes).cuda()
    correct = torch.zeros(config.num_classes).cuda()

    confidence = np.array([])
    pred_class = np.array([])
    true_class = np.array([])

    with torch.no_grad():
        end = time.time()
        for i, (images, target) in enumerate(val_loader):
            if config.gpu is not None:
                images = images.cuda(config.gpu, non_blocking=True)
            if torch.cuda.is_available():
                target = target.cuda(config.gpu, non_blocking=True)

            # compute output
            output = model(images)
            loss = criterion(output, target)

            # measure accuracy and record loss
            acc1, acc5 = accuracy(output, target, topk=(1,5))
            losses.update(loss.item(), images.size(0))
            top1.update(acc1[0], images.size(0))
            top5.update(acc5[0], images.size(0))

            _, predicted = output.max(1)
            target_one_hot = F.one_hot(target, config.num_classes)
            predict_one_hot = F.one_hot(predicted, config.num_classes)
            class_num = class_num + target_one_hot.sum(dim=0).to(torch.float)
            correct = correct + (target_one_hot + predict_one_hot == 2).sum(dim=0).to(torch.float)

            prob = torch.softmax(output, dim=1)
            confidence_part, pred_class_part = torch.max(prob, dim=1)
            confidence = np.append(confidence, confidence_part.cpu().numpy())
            pred_class = np.append(pred_class, pred_class_part.cpu().numpy())
            true_class = np.append(true_class, target.cpu().numpy())

            # measure elapsed time
            batch_time.update(time.time() - end)
            end = time.time()

            if i % config.print_freq == 0:
                progress.display(i, logger)

        acc_classes = correct / class_num
        head_acc = acc_classes[config.head_class_idx[0]:config.head_class_idx[1]].mean() * 100
        med_acc = acc_classes[config.med_class_idx[0]:config.med_class_idx[1]].mean() * 100
        tail_acc = acc_classes[config.tail_class_idx[0]:config.tail_class_idx[1]].mean() * 100
        logger.info(
            '* Acc@1 {top1.avg:.3f}% Acc@5 {top5.avg:.3f}% HAcc {head_acc:.3f}% MAcc {med_acc:.3f}% TAcc {tail_acc:.3f}%.'.format(
                top1=top1, top5=top5, head_acc=head_acc, med_acc=med_acc, tail_acc=tail_acc))

        cal = calibration(true_class, pred_class, confidence, num_bins=15)
        logger.info('* ECE   {ece:.3f}%.'.format(ece=cal['expected_calibration_error'] * 100))

    return top1.avg, cal['expected_calibration_error'] * 100


def save_checkpoint(state, is_best, model_dir):
    filename = model_dir + '/current.pth.tar'
    torch.save(state, filename)
    if is_best:
        shutil.copyfile(filename, model_dir + '/model_best.pth.tar')

def adjust_learning_rate(optimizer, epoch, args):
    """Sets the learning rate to the initial LR decayed by 10 every 30 epochs"""
    if args.cos:
        lr_min = 0
        lr_max = args.lr
        lr = lr_min + 0.5 * (lr_max - lr_min) * (1 + math.cos(epoch / args.num_epochs * 3.1415926535))
    else:
        epoch = epoch + 1
        if epoch <= 5:
            lr = args.lr * epoch / 5
        elif epoch > 180:
            lr = args.lr * 0.0001
        elif epoch > 160:
            lr = args.lr * 0.01
        else:
            lr = args.lr
    for param_group in optimizer.param_groups:
        param_group['lr'] = lr

if __name__ == '__main__':
    main()