import argparse
import pprint
import os
import time
import torch
import torch.nn.functional as F
from dataset import *
from utils import *
from models import build_model
from losses import *
from torch.optim.lr_scheduler import CosineAnnealingLR, StepLR
from torch.cuda.amp import autocast, GradScaler

def parse_args():
    parser = argparse.ArgumentParser(description='Learning with Noisy Labels')
    # dataset settings
    parser.add_argument('--root', default='../database', type=str, help='root directory')
    parser.add_argument('--dataset', default='CIFAR100', type=str, help='the dir of dataset')
    parser.add_argument('--noise_type', type=str, default='asymmetric', help='the noise type: clean, symmetric, pairflip, asymmetric')
    parser.add_argument('--noise_rate', type=float, default=0.4, help='the noise rate')
    # model settings
    parser.add_argument('--arch', default='cnn', type=str, choices=['cnn', 'resnet18', 'resnet34'], help='the network architecture')
    parser.add_argument('--mode', default='fnorm', type=str, choices=['linear', 'flinear', 'norm', 'fnorm'], help='the mode of the last linear layer')
    parser.add_argument('--need-linear', action='store_true', help='useful for discarding linear')
    parser.add_argument('--last-relu', default=False, action='store_true')
    parser.add_argument('--loss', default='ce', type=str)
    parser.add_argument('--scale', default=1.0, type=float)
    parser.add_argument('--temp', default=1.0, type=float)
    parser.add_argument('--temp-decay', default='inverse', type=str)
    parser.add_argument('--alpha', default=1.0, type=float)
    parser.add_argument('--beta', default=1.0, type=float)
    parser.add_argument('--scheduler', default='cos', type=str, help='the scheduler')
    # training settings
    parser.add_argument('--epochs', default=200, type=int, help='training epochs')
    parser.add_argument('--batch_size', type=int, default=128, help='batch size')
    parser.add_argument('--num_workers', type=int, default=10, help='the number of worker for loading data')
    parser.add_argument('--grad_clip', default=True, help='if use gradient clipping')
    parser.add_argument('--grad_bound', type=float, default=5., help='the gradient norm bound')
    parser.add_argument('--print-freq', default=50, type=int)

    # initialization settings
    parser.add_argument('--gpu-ids', default='0', type=str, help='id(s) for CUDA_VISIBLE_DEVICES')
    parser.add_argument('--resume', default='', type=str, help='path to latest checkpoint (default: none)')
    parser.add_argument('--deterministic', default=False, type=bool)
    parser.add_argument('--seed', default=0, type=int, help="random seed")
    parser.add_argument("--amp", default=False, action="store_true", help="use 16-bit (mixed) precision through NVIDIA apex AMP")
    parser.add_argument("--opt_level", type=str, default="O1", help="apex AMP optimization level selected in ['O0', 'O1', 'O2', and 'O3']."
                        "See details at https://nvidia.github.io/apex/amp.html")
    parser.add_argument("--local_rank", type=int, default=-1, help="For distributed training: local_rank")

    parser.add_argument('--save-path', default='RESULTS/LNL', type=str)
    parser.add_argument('--model-dir', default='model', type=str)
    parser.add_argument('--log-dir', default='log', type=str)
    parser.add_argument('--tag', default=None, type=str)

    args = parser.parse_args()
    return args

def set_config(args):
    scale = 1.0
    if args.loss == 'gce':
        scale = 2.0
    if args.temp_decay == 'exp':
        args.temp = scale * 5.0 / 2 ** (7 * args.noise_rate)
    else:
        args.temp = scale * 0.25 / (0.05 + args.noise_rate)
    if args.dataset == 'MNIST':
        args.weight_decay = 1e-3
        args.lr = 0.01
        args.epochs = 50
        args.num_classes = 10
        args.num_feats = 128
        args.arch = 'cnn'
    elif args.dataset == 'CIFAR10':
        args.weight_decay = 1e-4
        args.lr = 0.01
        args.num_classes = 10
        args.epochs = 120
        args.num_feats = 256
        args.arch = 'cnn'
    elif args.dataset == 'CIFAR100':
        args.weight_decay = 1e-5
        args.lr = 0.1
        args.epochs = 200
        args.num_classes = 100
        args.num_feats = 512
        args.arch = 'resnet34'
    elif args.dataset == 'WEBVISION':
        args.weight_decay = 3e-5
        args.lr = 0.4
        args.epochs = 250
        args.num_classes = 50
        args.num_feats = 2048
        args.batch_size = 256
        args.arch = 'resnet50'
        args.scheduler = 'step'
        args.grad_clip = False
    else:
        raise NotImplementedError
    
    if args.loss.lower() in ['nceandmae', 'nceandrce', 'nflandmae', 'nflandrce']:
        args.temp = 5.0
        if args.dataset == 'MNIST':
            args.alpha = 1.0
            args.beta = 100.0
        elif args.dataset == 'CIFAR10':
            args.alpha = 1.0
            args.beta = 1.0
        elif args.dataset == 'CIFAR100':
            args.alpha = 10.0
            args.beta = 0.1
    elif args.loss.lower() == 'sce':
        if args.dataset == 'MNIST':
            args.alpha = 0.01
            args.beta = 1.0
        elif args.dataset == 'CIFAR10':
            args.alpha = 0.1
            args.beta = 1.0
        elif args.dataset == 'CIFAR100':
            args.alpha = 6
            args.beta = 0.1


def main():
    args = parse_args()
    set_config(args)
    set_device(args)
    if args.tag is None:
        args.tag = '_'.join([args.dataset, args.noise_type, str(args.noise_rate)]) + '/' + '_'.join([str(op) for op in
            [args.loss, args.scale, args.temp, args.temp_decay, args.alpha, args.beta, args.mode, args.last_relu, args.lr, args.weight_decay, args.seed]])

    args, logger = create_logger(args)
    logger.info('\n' + pprint.pformat(args))

    if args.deterministic:
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        set_seed(args.seed)
    else:
        torch.backends.cudnn.benchmark = True

    data_loader = get_lnl_dataloader(args)
    train_loader, test_loader = data_loader['train_dataset'], data_loader['test_dataset']
    if args.dataset == 'Clothing1M':
        val_loader = data_loader['valid_dataset']

    model = build_model(args, mode=args.mode).to(args.device)
    optimizer = torch.optim.SGD(model.parameters(), lr=args.lr, momentum=0.9, weight_decay=args.weight_decay)
    if args.scheduler == 'step':
        scheduler = StepLR(optimizer, step_size=1, gamma=0.97)
    else:
        scheduler = CosineAnnealingLR(optimizer, T_max=args.epochs, eta_min=0.0)
    if args.amp:
        scaler = GradScaler()

    args.best_acc = 0.
    if args.resume:
        logger.info('==> Resuming from checkpoint..')
        assert os.path.isfile(args.resume), "Error: no checkpoint directory found!"
        args.out = os.path.dirname(args.resume)
        args.model_dir = os.path.dirname(args.resume)
        checkpoint = torch.load(args.resume)
        args.best_acc = checkpoint['best_acc']
        args.start_epoch = checkpoint['epoch']
        model.load_state_dict(checkpoint['state_dict'])
        optimizer.load_state_dict(checkpoint['optimizer'])
        scheduler.load_state_dict(checkpoint['scheduler'])

    logger.info("***** Running training *****")


    criterion = get_criterion(args, train_loader=train_loader)

    logger.info(f"  Task = {args.dataset}@{args.noise_type}@{args.noise_rate}")
    logger.info(f"  Num Epochs = {args.epochs}")
    logger.info(f"  Batch size per GPU = {args.batch_size}")
    logger.info(f"  Total train batch size = {args.batch_size*args.world_size}")
    for epoch in range(args.epochs):
        batch_time = AverageMeter('BatchTime', ':6.3f')
        data_time = AverageMeter('DataTime', ':6.3f')
        losses = AverageMeter('Loss', ':.3f')
        margins = AverageMeter('Margin', ':.3f')
        norms = AverageMeter('Norm', ':.3f')
        ratios = AverageMeter('Ratio', ':.3f')
        top1 = AverageMeter('Acc@1', ':6.3f')
        top5 = AverageMeter('Acc@5', ':6.3f')

        progress = ProgressMeter(int(len(train_loader.dataset)/args.batch_size),
                                 [batch_time, data_time, losses, margins, norms, ratios, top1, top5],
                                 prefix="Epoch: [{}]".format(epoch))

        model.train()
        end = time.time()
        for batch_idx, (batch_x, batch_y) in enumerate(train_loader):
            batch_x, batch_y = batch_x.to(args.device), batch_y.to(args.device)
            model.zero_grad()
            optimizer.zero_grad()

            if args.amp:
                with autocast():
                    logits = model(batch_x)
                    if args.loss == 'ce':
                        logits *= args.temp
                    loss = criterion(logits, batch_y)
            else:
                logits = model(batch_x)
                if args.loss == 'ce':
                    logits *= args.temp
                loss = criterion(logits, batch_y)

            if args.amp:
                scaler.scale(loss).backward()

                # Unscales the gradients of the optimizer's assigned params in-place
                scaler.unscale_(optimizer)
                # Since the gradients of optimizer's assigned params are unscaled, clips as usual:
                torch.nn.utils.clip_grad_norm_(model.parameters(), args.grad_bound)
                # optimizer;s gradients are already unscaled, so scaler.step does not unscale them,
                # although it still skips optimizer.step() if the gradients contain infs or NaNs.
                scaler.step(optimizer)
                # Updates the scale for next iteration.
                scaler.update()
            else:
                loss.backward()
                if args.grad_clip:
                    torch.nn.utils.clip_grad_norm_(model.parameters(), args.grad_bound)
                optimizer.step()

            losses.update(loss.item())

            try:
                margin, norm, ratio = model.margin()
            except:
                margin, norm, ratio = model.module.margin()

            margins.update(margin)
            norms.update(norm)
            ratios.update(ratio)
            acc1, acc5 = accuracy(logits, batch_y, topk=(1,5))
            top1.update(acc1[0], batch_x.size(0))
            top5.update(acc5[0], batch_x.size(0))

            batch_time.update(time.time() - end)
            end = time.time()

            if batch_idx % args.print_freq == 0:
                progress.display(batch_idx, logger)
        args.writer.add_scalar('train/lr', optimizer.param_groups[0]['lr'], epoch)
        args.writer.add_scalar('train/loss', losses.avg, epoch)
        args.writer.add_scalar('train/acc1', top1.avg, epoch)
        args.writer.add_scalar('train/acc5', top5.avg, epoch)
        scheduler.step()

        try:
            margin, norm, ratio = model.margin()
        except:
            margin, norm, ratio = model.module.margin()

        args.writer.add_scalar('network/margin', margin, epoch)
        args.writer.add_scalar('network/norm', norm, epoch)
        args.writer.add_scalar('network/ratio', ratio, epoch)

        acc1, acc5, sm = validate(args, test_loader, model)
        args.writer.add_scalar('test/acc1', acc1, epoch)
        args.writer.add_scalar('test/acc5', acc5, epoch)
        args.writer.add_scalar('test/sample_margin', sm, epoch)

        is_best = acc1 > args.best_acc
        args.best_acc = max(acc1, args.best_acc)
        save_checkpoint({
            'epoch': epoch + 1,
            'state_dict': model.state_dict(),
            'acc': acc1,
            'best_acc': args.best_acc,
            'optimizer': optimizer.state_dict(),
            'scheduler': scheduler.state_dict(),
        }, is_best, args.model_dir)

@torch.no_grad()
def validate(args, loader, model):
    top1 = AverageMeter('Acc@1', ':6.3f')
    top5 = AverageMeter('Acc@5', ':6.3f')
    sms = AverageMeter('SampleMargin', ':.4f')
    sample_margin = SampleMarginLoss()
    model.eval()
    for inputs, targets in loader:
        inputs, targets = inputs.to(args.device), targets.to(args.device)
        try:
            embedding = model.get_feature(inputs)
            outputs = model.linear(embedding)
            weight = model.linear.weight
        except:
            embedding = model.module.get_feature(inputs)
            outputs = model.module.linear(embedding)
            weight = model.module.linear.weight
        embedding_norm = F.normalize(embedding, dim=1)
        weight_norm = F.normalize(weight, dim=1)
        outputs_norm = F.linear(embedding_norm, weight_norm)
        prec1, prec5 = accuracy(outputs, targets, topk=(1, 5))

        sm = sample_margin(outputs_norm, targets)
        top1.update(prec1[0], inputs.size(0))
        top5.update(prec5[0], inputs.size(0))
        sms.update(sm.item(), inputs.size(0))
    logger.info("\nValidation: top-1 acc: {:.2f}, top-5 acc: {:.2f}, sample margin: {:.3f}\n".format(top1.avg, top5.avg, -sms.avg))
    return top1.avg, top5.avg, -sms.avg


if __name__ == "__main__":
    main()