
import torch
import torch.nn
from .arcface_loss import *
from .cosface_loss import *
from .focal_loss import *
from .ldam_loss import *
from .largest_margin_softmax_loss import *
from .normface_loss import *
from .sample_margin_loss import *
from .large_margin_softmax_loss import *
from .sphereface_loss import *
from .balanced_softmax import *
from .sample_logit_loss import *
from .robust_losses import *


def get_criterion(args, train_loader=None):
    loss = args.loss.lower()
    if loss == 'ce':
        return nn.CrossEntropyLoss()
    elif loss == 'mae':
        return MAELoss(num_classes=args.num_classes, scale=args.scale, temp=args.temp)
    elif loss == 'gce':
        return GCELoss(num_classes=args.num_classes, q=0.7, temp=args.temp)
    elif loss == 'sample_logit':
        return NegSampleLogitLoss(scale=args.scale)
    elif loss == 'sample_margin':
        return SampleMarginLoss()
    elif loss == 'normface':
        return NormFaceLoss(temp=args.temp)
    elif loss == 'nceandsl':
        return NCEandSL(alpha=args.alpha, beta=args.beta, num_classes=args.num_classes, temp=args.temp)
    elif loss == 'nceandmae':
        return NCEandMAE(alpha=args.alpha, beta=args.beta, num_classes=args.num_classes, temp=args.temp)
    elif loss == 'nslandmae':
        return NSLandMAE(alpha=args.alpha, beta=args.beta, num_classes=args.num_classes, temp=args.temp)
    elif loss == 'nlnl':
        return NLNL(train_loader, num_classes=args.num_classes, temp=args.temp)
    elif loss == 'focal':
        return FocalLoss()
    elif loss == 'sce':
        return SCELoss(num_classes=args.num_classes, a=args.alpha, b=args.beta, temp=args.temp)
    elif loss == 'nceandmae':
        return NCEandMAE(alpha=args.alpha, beta=args.beta, num_classes=args.num_classes, temp=args.temp)
    elif loss == 'nceandrce':
        return NCEandRCE(alpha=args.alpha, beta=args.beta, num_classes=args.num_classes, temp=args.temp)
    elif loss == 'nflandmae':
        return NFLandMAE(alpha=args.alpha, beta=args.beta, num_classes=args.num_classes, temp=args.temp)
    elif loss == 'nflandrce':
        return NFLandRCE(alpha=args.alpha, beta=args.beta, num_classes=args.num_classes, temp=args.temp)
    elif loss == 'agce':
        return AGCELoss(num_classes=args.num_classes, a=4, q=0.2)
    elif loss == 'aul':
        return AUELoss(num_classes=args.num_classes, a=3, q=0.1)
    else:
        raise NotImplementedError


class NCEandRCE(nn.Module):
    def __init__(self, alpha=1., beta=1., num_classes=10, temp=1.0):
        super(NCEandRCE, self).__init__()
        self.num_classes = num_classes
        self.nce = NCELoss(num_classes=num_classes, scale=alpha, temp=temp)
        self.rce = RCELoss(num_classes=num_classes, scale=beta, temp=temp)

    def forward(self, pred, labels):
        return self.nce(pred, labels) + self.rce(pred, labels)

class NCEandMAE(nn.Module):
    def __init__(self, alpha=1., beta=1., num_classes=10, temp=1.0):
        super(NCEandMAE, self).__init__()
        self.num_classes = num_classes
        self.nce = NCELoss(num_classes=num_classes, scale=alpha, temp=temp)
        self.mae = MAELoss(num_classes=num_classes, scale=beta, temp=temp)

    def forward(self, pred, labels):
        return self.nce(pred, labels) + self.mae(pred, labels)


class NFLandRCE(torch.nn.Module):
    def __init__(self, alpha=1., beta=1., num_classes=10, gamma=0.5, temp=1.0):
        super(NFLandRCE, self).__init__()
        self.num_classes = num_classes
        self.nfl = NormalizedFocalLoss(gamma=gamma, num_classes=num_classes, scale=alpha, temp=temp)
        self.rce = RCELoss(num_classes=num_classes, scale=beta, temp=temp)

    def forward(self, pred, labels):
        return self.nfl(pred, labels) + self.rce(pred, labels)


class NFLandMAE(torch.nn.Module):
    def __init__(self, alpha=1., beta=1., num_classes=10, gamma=0.5, temp=1.0):
        super(NFLandMAE, self).__init__()
        self.num_classes = num_classes
        self.nfl = NormalizedFocalLoss(gamma=gamma, num_classes=num_classes, scale=alpha, temp=temp)
        self.mae = MAELoss(num_classes=num_classes, scale=beta, temp=temp)

    def forward(self, pred, labels):
        return self.nfl(pred, labels) + self.mae(pred, labels)


class NCEandAGCE(torch.nn.Module):
    def __init__(self, alpha=1., beta = 1., num_classes=10, a=3, q=1.5, temp=1.0):
        super(NCEandAGCE, self).__init__()
        self.num_classes = num_classes
        self.nce = NCELoss(num_classes=num_classes, scale=alpha, temp=temp)
        self.agce = AGCELoss(num_classes=num_classes, a=a, q=q, scale=beta, temp=temp)

    def forward(self, pred, labels):
        return self.nce(pred, labels) + self.agce(pred, labels)


class NCEandAUE(torch.nn.Module):
    def __init__(self, alpha=1., beta=1., num_classes=10, a=6, q=1.5, temp=1.0):
        super(NCEandAUE, self).__init__()
        self.num_classes = num_classes
        self.nce = NCELoss(num_classes=num_classes, scale=alpha, temp=temp)
        self.aue = AUELoss(num_classes=num_classes, a=a, q=q, scale=beta, temp=temp)

    def forward(self, pred, labels):
        return self.nce(pred, labels) + self.aue(pred, labels)

class NCEandAEL(torch.nn.Module):
    def __init__(self, alpha=1., beta=4., num_classes=10, a=2.5, temp=1.0):
        super(NCEandAEL, self).__init__()
        self.num_classes = num_classes
        self.nce = NCELoss(num_classes=num_classes, scale=alpha, temp=temp)
        self.aue = AExpLoss(num_classes=num_classes, a=a, scale=beta, temp=temp)

    def forward(self, pred, labels):
        return self.nce(pred, labels) + self.aue(pred, labels)


class NFLandAGCE(torch.nn.Module):
    def __init__(self, alpha=1., beta = 1., num_classes=10, a=3, q=2, temp=1.0):
        super(NFLandAGCE, self).__init__()
        self.num_classes = num_classes
        self.nce = NormalizedFocalLoss(num_classes=num_classes, scale=alpha, temp=temp)
        self.agce = AGCELoss(num_classes=num_classes, a=a, q=q, scale=beta, temp=temp)

    def forward(self, pred, labels):
        return self.nce(pred, labels) + self.agce(pred, labels)


class NFLandAUE(torch.nn.Module):
    def __init__(self, alpha=1., beta = 1., num_classes=10, a=1.5, q=0.9, temp=1.0):
        super(NFLandAUE, self).__init__()
        self.num_classes = num_classes
        self.nce = NormalizedFocalLoss(num_classes=num_classes, scale=alpha, temp=temp)
        self.aue = AUELoss(num_classes=num_classes, a=a, q=q, scale=beta, temp=temp)

    def forward(self, pred, labels):
        return self.nce(pred, labels) + self.aue(pred, labels)


class NFLandAEL(torch.nn.Module):
    def __init__(self, alpha=1., beta = 1., num_classes=10, a=3, temp=1.0):
        super(NFLandAEL, self).__init__()
        self.num_classes = num_classes
        self.nce = NormalizedFocalLoss(num_classes=num_classes, scale=alpha, temp=temp)
        self.ael = AExpLoss(num_classes=num_classes, a=a, scale=beta, temp=temp)

    def forward(self, pred, labels):
        return self.nce(pred, labels) + self.ael(pred, labels)

class ANLandRCE(torch.nn.Module):
    def __init__(self, alpha=1., beta=1., num_classes=10, a=2, p=0.7, temp=1.0):
        super(ANLandRCE, self).__init__()
        self.num_classes = num_classes
        self.anl = ANormLoss(num_classes=num_classes, a=a, p=p, scale=alpha, temp=temp)
        self.rce = RCELoss(num_classes=num_classes, scale=beta, temp=temp)

    def forward(self, pred, labels):
        return self.anl(pred, labels) + self.rce(pred, labels)

class NCEandANL(torch.nn.Module):
    def __init__(self, alpha=1., beta=1., num_classes=10, a=2, p=0.7, temp=1.0):
        super(NCEandANL, self).__init__()
        self.num_classes = num_classes
        self.nce = NCELoss(num_classes=num_classes, scale=alpha, temp=temp)
        self.anl = ANormLoss(num_classes=num_classes, a=a, p=p, scale=beta, temp=temp)

    def forward(self, pred, labels):
        return self.nce(pred, labels) + self.anl(pred, labels)

class NCEandSL(torch.nn.Module):
    def __init__(self, alpha=0.5, beta=0.5, num_classes=10, temp=1.0):
        super(NCEandSL, self).__init__()
        self.nce = NCELoss(num_classes=num_classes, scale=alpha, temp=temp)
        self.nsl = NegSampleLogitLoss(scale=beta)

    def forward(self, logits, labels):
        return self.nce(logits, labels) + self.sl(logits, labels)

class NSLandMAE(torch.nn.Module):
    def __init__(self, alpha=0.5, beta=0.5, num_classes=10, temp=1.0):
        super(NSLandMAE, self).__init__()
        self.sl = NegSampleLogitLoss(scale=alpha)
        self.mae = MAELoss(num_classes=num_classes, scale=beta, temp=temp)

    def forward(self, logits, labels):
        return self.sl(logits, labels) + self.mae(logits, labels)