import math
import numpy as np
import scipy.special
import torch
import torch.nn as nn
import torch.nn.functional as F

class LargeMarginSoftmaxLoss(nn.Module):
    def __init__(self, margin=4, weight=None):
        super(LargeMarginSoftmaxLoss, self).__init__()
        self.eps = 1e-7
        self.margin = int(margin)
        self.divisor = math.pi / self.margin
        self.beta = 100
        self.beta_min = 0
        self.scale = 0.99
        self.max_n = self.margin // 2
        self.n_range = torch.tensor([n for n in range(0, self.max_n + 1)])
        self.margin_choose_n = torch.tensor(
            [scipy.special.binom(self.margin, 2 * n) for n in self.n_range]
        )
        self.cos_powers = torch.tensor([self.margin - (2 * n) for n in self.n_range])
        self.alternating = torch.tensor([(-1) ** n for n in self.n_range])
        self.weight = weight

    def cast_device(self, device):
        self.n_range = self.n_range.to(device)
        self.margin_choose_n = self.margin_choose_n.to(device)
        self.cos_powers = self.cos_powers.to(device)
        self.alternating = self.alternating.to(device)

    def get_cos_with_margin(self, cosine):
        cosine = cosine.unsqueeze(1)
        cos_powered = cosine ** self.cos_powers
        sin_powered = (1 - cosine ** 2) ** self.n_range
        terms = self.alternating * self.margin_choose_n * cos_powered * sin_powered
        return torch.sum(terms, dim=1)


    def get_cosine(self, embeddings, weights):
        norm_weights = F.normalize(weights, dim=1)
        norm_embeddings = F.normalize(embeddings, dim=1)
        return F.linear(norm_embeddings, norm_weights)

    def scale_logits(self, logits, embeddings, weights=None):
        embedding_norms = torch.norm(embeddings, p=2, dim=1)
        weight_norms = torch.norm(weights, p=2, dim=1)
        product_of_magnitudes = weight_norms.unsqueeze(0) * embedding_norms.unsqueeze(1)
        return logits * product_of_magnitudes * self.scale

    def modify_cosine_of_target_classes(self, cosine_of_target_classes):
        angles = torch.acos(torch.clamp(cosine_of_target_classes, -1 + self.eps, 1 - self.eps))
        cos_with_margin = self.get_cos_with_margin(cosine_of_target_classes)
        k = (angles / self.divisor).floor().detach()
        return ((-1) ** k) * cos_with_margin - (2 * k)


    def forward(self, weights, embeddings, labels):
        """
        :param weights: k x d
        :param embeddings: N x d
        :param labels
        :return:
        """
        beta = max(self.beta, self.beta_min)
        self.cast_device(weights.device)
        mask = F.one_hot(labels, weights.size()[0]).float().to(weights.device)
        cosine = self.get_cosine(embeddings, weights)
        cosine_of_target_classes = cosine[mask == 1]
        modified_cosine_of_target_classes = self.modify_cosine_of_target_classes(cosine_of_target_classes)
        modified_cosine_of_target_classes_beta = (modified_cosine_of_target_classes + beta * cosine_of_target_classes) / (1 + beta)
        diff = (modified_cosine_of_target_classes_beta - cosine_of_target_classes).unsqueeze(1)
        logits = cosine + (mask * diff)
        logits = self.scale_logits(logits, embeddings, weights)
        self.beta *= self.scale
        return F.cross_entropy(logits, labels)


if __name__ == '__main__':
    torch.manual_seed(123)
    criterion = LargeMarginSoftmaxLoss()
    x = torch.randn(5, 4)
    w = torch.randn(3, 4)
    print(x)
    print(w)
    y = torch.LongTensor([0, 0, 1, 2, 1])
    print(criterion(w, x, y))