import torch
import torch.nn as nn
import torch.nn.functional as F
from .basic import*

class ConvBrunch(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size=3):
        super(ConvBrunch, self).__init__()
        padding = (kernel_size - 1) // 2
        self.conv = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size, padding=padding),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True)
        )
    
    def forward(self, x):
        return self.conv(x)


class CNN(Basic):
    """ CNN for MNIST and CIFAR-10 """
    def __init__(self, dataset='CIFAR10', mode='linear', weight=None, last_relu=False):
        num_classes = 10
        if dataset == 'MNIST':
            num_feats = 128
        elif dataset == 'CIFAR10':
            num_feats = 256
        else:
            raise NotImplementedError
        super(CNN, self).__init__(num_feats, num_classes, mode=mode, weight=weight)
        self.dataset = dataset
        self.last_relu = last_relu
        if dataset == "CIFAR10":
            self.block1 = nn.Sequential(
                ConvBrunch(3, 64, 3),
                ConvBrunch(64, 64, 3),
                nn.MaxPool2d(kernel_size=2, stride=2)
            )
            self.block2 = nn.Sequential(
                ConvBrunch(64, 128, 3),
                ConvBrunch(128, 128, 3),
                nn.MaxPool2d(kernel_size=2, stride=2),
                ConvBrunch(128, 196, 3),
                ConvBrunch(196, 196, 3),
                nn.MaxPool2d(kernel_size=2, stride=2)
            )
            self.fc1 = nn.Sequential(
                nn.Linear(4 * 4 * 196, 256),
                nn.BatchNorm1d(256),
            )
            self.fc_size = 4 * 4 * 196
        elif dataset == "MNIST":
            self.block1 = nn.Sequential(
                ConvBrunch(1, 32, 3),
                nn.MaxPool2d(kernel_size=2, stride=2)
            )
            self.block2 = nn.Sequential(
                ConvBrunch(32, 64, 3),
                nn.MaxPool2d(kernel_size=2, stride=2)
            )
            self.fc1 = nn.Sequential(
                nn.Linear(64 * 7 * 7, 128),
                nn.BatchNorm1d(128),
            )
            self.fc_size = 64 * 7 * 7
        
        self._reset_params()
        
    
    def _reset_params(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_uniform_(m.weight, mode='fan_in', nonlinearity='relu')
            elif isinstance(m, nn.Linear):
                nn.init.xavier_uniform_(m.weight)
    
    def get_feature(self, x):
        x = self.block1(x)
        x = self.block2(x)
        x = x.view(-1, self.fc_size)
        x = self.fc1(x)
        if self.last_relu:
            x = F.relu(x)
        return x
    
