
import torch
import math
import torch.nn as nn
import torch.nn.functional as F

__all__ = ['NormLinear', 'FLinear', 'FNormLinear', 'InitLinear', 'Classifier']


class NormLinear(nn.Module):
    """
        L2-normalization for both prototypes and inputs
    """
    __constants__ = ['bias', 'in_features', 'out_features']
    def __init__(self, in_features, out_features, bias=False):
        super(NormLinear, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = nn.Parameter(torch.Tensor(out_features, in_features))
        self.reset_parameters()

    def reset_parameters(self):
        nn.init.kaiming_uniform_(self.weight.data, a=math.sqrt(5))
        # torch.nn.init.normal_(self.weight)

    def forward(self, input):
        input = F.normalize(input, dim=1)
        weight = F.normalize(self.weight, dim=1)
        return F.linear(input, weight)

    def extra_repr(self) -> str:
        return 'in_features={}, out_features={}'.format(
            self.in_features, self.out_features
        )

class FLinear(nn.Module):
    """
        The linear layer with fixed prototypes
    """
    __constants__ = ['bias', 'in_features', 'out_features']
    def __init__(self, in_features, out_features, weight):
        super(FLinear, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = nn.Parameter(weight)
        self.weight.requires_grad = False

    def forward(self, input):
        return F.linear(input, self.weight)

    def extra_repr(self) -> str:
        return 'in_features={}, out_features={}'.format(
            self.in_features, self.out_features
        )

class InitLinear(nn.Module):
    __constants__ = ['bias', 'in_features', 'out_features']
    def __init__(self, in_features, out_features, weight):
        super(InitLinear, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = nn.Parameter(weight)

    def forward(self, input):
        return F.linear(input, self.weight)

    def extra_repr(self) -> str:
        return 'in_features={}, out_features={}'.format(
            self.in_features, self.out_features
        )

class InitNormLinear(nn.Module):
    __constants__ = ['bias', 'in_features', 'out_features']
    def __init__(self, in_features, out_features, weight):
        super(InitNormLinear, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = nn.Parameter(weight)

    def forward(self, input):
        input = F.normalize(input, dim=1)
        weight = F.normalize(self.weight, dim=1)
        return F.linear(input, weight)

    def extra_repr(self) -> str:
        return 'in_features={}, out_features={}'.format(
            self.in_features, self.out_features
        )


class FNormLinear(nn.Module):
    """
        L2-normalization for both prototypes and inputs when prototypes are fixed.
    """
    __constants__ = ['bias', 'in_features', 'out_features']
    def __init__(self, in_features, out_features, weight):
        super(FNormLinear, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = nn.Parameter(weight)
        self.weight.requires_grad = False

    def forward(self, input):
        input = F.normalize(input, dim=1)
        weight = F.normalize(self.weight, dim=1)
        return F.linear(input, weight)

    def extra_repr(self) -> str:
        return 'in_features={}, out_features={}'.format(
            self.in_features, self.out_features
        )


class Classifier(nn.Module):
    def __init__(self, in_features, num_classes, mode='none', weight=None, bias=False, need_avg=False, need_pre=False):
        super(Classifier, self).__init__()
        self.need_avg = need_avg
        self.need_pre = need_pre
        if self.need_avg:
            self.avgpool = nn.AdaptiveAvgPool2d((1, 1))

        if self.need_pre:
            self.pre_linear = nn.Linear(in_features, in_features)

        self.mode = 'norm' if mode.endswith('norm') else 'linear'
        if mode == 'linear' and weight is None:
            self.linear = nn.Linear(in_features, num_classes, bias=bias)
            nn.init.kaiming_normal_(self.linear.weight)
        elif mode == 'norm' and weight is None:
            self.linear = NormLinear(in_features, num_classes, bias=bias)
        elif mode == 'fixnorm' and weight is not None:
            self.linear = FNormLinear(in_features, num_classes, weight)
        elif mode == 'fixlinear' and weight is not None:
            self.linear = FLinear(in_features, num_classes, weight)
        elif mode == 'initlinear' and weight is not None:
            self.linear = InitLinear(in_features, num_classes, weight)
        elif mode == 'initnorm' and weight is not None:
            self.linear = InitNormLinear(in_features, num_classes, weight)
        else:
            raise NotImplementedError

    def change(self, mode='none'):
        self.mode = mode

    def forward(self, x):
        if self.need_avg:
            x = self.avgpool(x)
            x = torch.flatten(x, 1)
        if self.need_pre:
            x = self.pre_linear(x)

        if self.mode == 'norm':
            input = F.normalize(x, dim=1)
            weight = F.normalize(self.linear.weight, dim=1)
            return F.linear(input, weight)
        else:
            return F.linear(x, self.linear.weight)

    def margin(self):
        assert hasattr(self, 'linear')
        weight = self.linear.weight
        norm = torch.sqrt(torch.sum(weight ** 2, dim=1))
        min_norm = torch.min(norm)
        ratio = torch.max(norm) / torch.min(norm)

        tmp = F.normalize(weight, dim=1)
        similarity = torch.matmul(tmp, tmp.transpose(1, 0)) - 2 * torch.eye(tmp.size(0), device=weight.device)
        return torch.acos(torch.max(similarity)).item() / math.pi * 180, min_norm, ratio.item()