
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init
from .basic import *

__all__ = ['cifar_resnet32']

def _weights_init(m):
    if isinstance(m, nn.Linear) or isinstance(m, nn.Conv2d):
        init.kaiming_normal_(m.weight)

class LambdaLayer(nn.Module):

    def __init__(self, lambd):
        super(LambdaLayer, self).__init__()
        self.lambd = lambd

    def forward(self, x):
        return self.lambd(x)


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, in_planes, planes, stride=1, option='A', use_relu=True):
        super(BasicBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.use_relu = use_relu

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != planes:
            if option == 'A':
                """
                For CIFAR10 ResNet paper uses option A.
                """
                self.shortcut = LambdaLayer(lambda x:
                                            F.pad(x[:, :, ::2, ::2], (0, 0, 0, 0, planes//4, planes//4), "constant", 0))
            elif option == 'B':
                self.shortcut = nn.Sequential(
                     nn.Conv2d(in_planes, self.expansion * planes, kernel_size=1, stride=stride, bias=False),
                     nn.BatchNorm2d(self.expansion * planes)
                )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        if self.use_relu:
            out = F.relu(out)
        return out

class ResNet_s(Basic):
    def __init__(self, block, num_blocks, num_classes=10, num_feats=64, mode='', last_relu=True, need_linear=True, weight=None):
        super(ResNet_s, self).__init__(num_feats, num_classes, mode=mode, need_linear=need_linear, weight=weight)
        self.in_planes = 16

        self.conv1 = nn.Conv2d(3, 16, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(16)
        self.layer1 = self._make_layer(block, 16, num_blocks[0], stride=1)
        self.layer2 = self._make_layer(block, 32, num_blocks[1], stride=2)
        option = 'A' if num_feats == 64 else 'B'
        self.layer3 = self._make_layer(block, 64, num_blocks[2], stride=2, option=option, num_feats=num_feats, last_relu=last_relu)
        self.apply(_weights_init)

    def _make_layer(self, block, planes, num_blocks, stride, option='A', num_feats=64, last_relu=True):
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []
        for i, stride in enumerate(strides):
            if i == num_blocks - 1 and not last_relu:
                layers.append(block(self.in_planes, num_feats, stride, option=option, use_relu=last_relu))
            else:
                layers.append(block(self.in_planes, planes, stride))
            self.in_planes = planes * block.expansion

        return nn.Sequential(*layers)

    def get_feature(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = F.avg_pool2d(out, out.size()[3])
        out = out.view(out.size(0), -1)
        return out

def cifar_resnet32(num_classes=10, num_feats=64, mode='', last_relu=True, need_linear=True, weight=None):
    return ResNet_s(BasicBlock, [5, 5, 5], num_classes=num_classes, num_feats=num_feats, mode=mode, last_relu=last_relu, need_linear=need_linear, weight=weight)

def count_parameters(model):
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


if __name__ == "__main__":
    net = cifar_resnet32(num_classes=10, mode='')
    print(count_parameters(net))